/*
//
//                  INTEL CORPORATION PROPRIETARY INFORMATION
//     This software is supplied under the terms of a license agreement or
//     nondisclosure agreement with Intel Corporation and may not be copied
//     or disclosed except in accordance with the terms of that agreement.
//          Copyright(c) 2007-2008 Intel Corporation. All Rights Reserved.
//
//     Intel(R) Integrated Performance Primitives Data Compression Domain-based GZIP
//
//  By downloading and installing this sample, you hereby agree that the
//  accompanying Materials are being provided to you under the terms and
//  conditions of the End User License Agreement for the Intel(R) Integrated
//  Performance Primitives product previously accepted by you. Please refer
//  to the file ippEULA.rtf located in the root directory of your Intel(R) IPP
//  product installation for more information.
//
*/
#ifndef IPP_GZIP_H
#define IPP_GZIP_H

#if defined(_OPENMP)
   #if !defined(GZIP_PARALLEL)
       #warning WARNING: Compiling file in SINGLE-THREAD mode. GZIP_PARALLEL must be defined (-DGZIP_PARALLEL) to have true multi-thread capabilities.
   #endif
#endif

#include <stdio.h>
#include <sys/stat.h>
#ifdef _WIN32
    #include <windows.h>
#endif
#ifdef GZIP_PTHREADS
    #include <pthread.h>
#endif
#if !defined _WIN32
        #include <sys/mman.h>
        #include <unistd.h>
        #include <fcntl.h>
#endif
#include "ippcore.h"

typedef unsigned char UBYTE;
#if !defined(_WIN32)
        typedef size_t DWORD;
    typedef unsigned short BOOL;
#endif

/*! \brief Ascii text mode selected */
extern short opt_ascii;
/*! \brief Print uncompressed output to stdout */
extern short opt_to_stdout;
/*! \brief Decompress the file(s) */
extern short opt_decompress;
/*! \brief NOT IMPLEMENTED: encrypt the compressed stuff */
extern short opt_encrypt;
extern short opt_force;
extern short opt_help;
extern short opt_pkzip;
extern short opt_list;
extern short opt_license;
extern short opt_force_dynamic;
/*! \brief Don't save/restore original name */
extern short opt_noname;
extern short opt_name;
extern short opt_quiet;
extern short opt_reqursive;
extern short opt_new_suffix;
extern short opt_test;
extern short opt_notime;
extern short opt_verbose;
extern short opt_version;
extern short opt_compress_fast;
extern short opt_compress_best;
extern short opt_lzw_compat;
extern short opt_bits;
extern short opt_rsyncable;
extern char *prog_name;
extern char *suffix;
extern int num_threads;
extern short from_stdin;

/* Input/output options */
extern short opt_io_read_method;
extern char *ptr_read_method;
extern short opt_io_write_method;
extern char *ptr_write_method;
extern short opt_io_read_buffer;
extern size_t INBUF_SIZE;
extern char *ptr_read_buffer;
extern short opt_io_write_buffer;
extern size_t OUTBUF_SIZE;
extern char *ptr_write_buffer;
extern short opt_perf_stat;


extern short opt_dbg_multithreads;
extern char *opt_num_threads_forced;
extern short opt_dbg_slice_size;
extern char *opt_slice_size_array;
extern long opt_slice_size;

extern short opt_debug;

#ifdef GZIP_PTHREADS
extern pthread_cond_t condition_cond;
extern pthread_mutex_t condition_mutex;
extern pthread_t f_pthread;                                     /* Finishing pthread identifier */
#endif

/* File i/o interface */
enum TYPE_GZIP_FILE_MODE {
    GZIP_FILE_READ,
    GZIP_FILE_WRITE
};

#define METHOD_LOW     0
#define METHOD_MMAP    1

typedef enum TYPE_GZIP_FILE_MODE GZIP_FILE_MODE;

struct ipp_gzip_file {
    GZIP_FILE_MODE mode;
    size_t bytes_out;   /* Total number of bytes written */
    size_t bytes_in;    /* Total number of bytes read */
    int eof_condition;  /* 1 if EOF met */
    int fd;             /* File descriptor number */
    /* The following fields are for file mapping and will not be used for read file operations */
    DWORD allocation_granularity;  /* Memory chunk size */
    size_t fileOffset;               /* This file offset must be multiple of dwAllocationGranularity */
    size_t curFilePosition;          /* This is a stream-like position pointer to compute offsets */
    size_t size;                 /* File size for input files */
    size_t uncompressed_size;
    UBYTE *pmap;                    /* Void pointer to mapped memory */
    UBYTE *pview_pos;              /* Current read position in view window */
    size_t valid_bytes;               /* Number of bytes is view window to read */
    size_t bufOffset;                /* Next detalization of offset with file. Thus, next read/write will be at fileOffset+bufOffset */
    #ifdef _WIN32
        HANDLE hFile;                   /* File handle for using in map object */
        HANDLE hMap;                    /* Fime map handle */
    #else                               /* Linux */
        #define MAP_PAGE_SIZE (64*1024)
        size_t maplen;
#endif
};

typedef struct ipp_gzip_file GZIP_FILE;


struct ipp_gzip_view {
    GZIP_FILE *file_ptr;
    UBYTE *pmap;
    UBYTE *pview_pos;
    size_t fileOffset;
    size_t valid_bytes;
    size_t bytes_out;
    size_t size;
    size_t curFilePosition;
    #ifndef _WIN32
        size_t maplen;
    #endif
};
typedef struct ipp_gzip_view GZIP_VIEW;

#if defined (__MACH__)
        #define ZOPEN zzopen
#else
        #define ZOPEN zzopen
#endif
GZIP_FILE* ZOPEN(char*, GZIP_FILE_MODE, size_t);    /* Function to open file for read or write */
int zclose(GZIP_FILE*);                             /* Close file function */
size_t zread(GZIP_FILE*, UBYTE*, size_t);            /* Read specified number of bytes to specified address */
size_t zwrite(GZIP_FILE*, UBYTE*, size_t);           /* Write buffer */
int zseek(GZIP_FILE*, long, int);                   /* Move file pointer to specified position */
int zeof(GZIP_FILE*);                               /* Check if EOF */
long ztell(GZIP_FILE*);                             /* Return current file position */
void zcopytime(char*, char*, long);                 /* Copy date/time information from one file to another */
GZIP_FILE* zopen_stdin(void);                           /* Open standard input stream */
GZIP_FILE* zopen_stdout(void);                          /* Open standard output stream */
size_t zgetsize(GZIP_FILE*);                        /* Get file size */
DWORD zsizeuncompressed(GZIP_FILE*);                /* Read and return umcompressed file size from GZIP file */
int zfilemap(GZIP_FILE*, DWORD);
void ztruncate(GZIP_FILE*);
GZIP_VIEW* zopenview(GZIP_FILE*, DWORD);
size_t zwriteview(GZIP_VIEW*, UBYTE*, size_t);
void zcloseview(GZIP_VIEW*);
BOOL zextend(GZIP_FILE*, DWORD);

/* Global function prototypes */
void print_version(void);
int get_options(int, char **);
void help(void);
void warn_message(char*);
char *get_program_name(char*);
void decompress(char*);
void compress(char*);
void *deflate_mt(char*);     /* Deflate multithreaded version */
#ifdef GZIP_PTHREADS
void *deflate_st(void*);
#else
void *deflate_st(char*);     /* Deflate singlethreaded version */
#endif
void inflate_mt(char*);
void inflate_st(char*);
void error_file(char*);     /* Print file error message and exit */
void *zalloc(size_t);  /* Speaking allocation function */
void zfree(void*);          /* Speaking memfree function */
void create_output_name(char *, char *);
#if _MSC_VER >= 1400  && !defined(_AMD64_)
UBYTE *create_gzip_hdr(UBYTE*, char*, struct _stat *, int);
#else
UBYTE *create_gzip_hdr(UBYTE*, char*, struct stat *, int);
#endif
void get_base_name(char*, char*);
int get_nosuf_name(char*, char*);
void restore_output_name(char*, char*, char*);
void print_ipp_status(IppStatus);
void print_progname(void);
GZIP_FILE *out_open(char*, size_t);
void show_list(char*, GZIP_FILE*,long);
char *mk_chunk_name(char*, int, char*);
int remove_tmp_file(char*, int);
#ifdef GZIP_PTHREADS
void *compress_dir_st(void*);
#else
void *compress_dir_st(char*);
#endif
int get_cpu_count(void);
int known_suffix(char*);
int is_file(int);
void error_system(char*);
void print_io_methods(void);

/* Global definitions */
/*! \brief Max threads to allow */
#define MAX_THREADS 32
/*! \brief Minimum file length to use chunks */
#define MIN_LENGTH_TO_SLICE (1024 * 256)
/*! \brief GZIP magic byte 1 */
#define GZIP_HDR_1 0x1F
/*! \brief GZIP magic byte 2 */
#define GZIP_HDR_2 0x8B
/*! \brief DEFLATE byte in header*/
#define DEFLATE 0x08
/* Flag byte: definition of bits */
/*! \brief Compressed file is text */
#define FTEXT 0x01
/*! \brief CRC16 is stored */
#define FHCRC16 0x02
/*! \brief Extra fields are present */
#define FEXTRA 0x04
/*! \brief Filename stored in gzip header */
#define FNAME 0x08
/*! \brief Comment field is present. In multi-chunk mode comment contains chunk params */
#define FCOMMENT 0x10
/*! \brief Extra flag: max compression */
#define XFL_BEST 0x02
/*! \brief Extra flag: best compression */
#define XFL_FAST 0x04
/*! \brief NTFS-based file system */
#define STORED_BUF_SIZE (1024*70)
/*! \brief Buffer size for stored block saving */
#ifdef _WIN32
#define OS_NTFS 0x0B
#define OS_BYTE OS_NTFS
#else
#define OS_UNIX 0x03
#define OS_BYTE OS_UNIX
#endif
/*! \brief Standard GZIP header size */
#define GZIP_HEADER_SIZE 10
/*! \brief Specific IPP GZIP comment field header */
#define COMMENT_MAGIC_HEADER 0xC0

#ifdef _WIN32
    #define MAX_NAME _MAX_PATH
#else
    #include <limits.h>
    #define MAX_NAME PATH_MAX
#endif

/*! \brief Put one byte to output buffer */
#define PUT_BYTE(o, b) *o++ = (UBYTE)(b)
/*! \brief Put long (4-byte) value to output buffer */
#define PUT_LONG(o, L) { PUT_BYTE(o, L & 0xFF); PUT_BYTE(o, (L>>8)&0xFF); PUT_BYTE(o, (L>>16)&0xFF); PUT_BYTE(o, (L>>24)&0xFF); }

/*! \brief Get one byte from input buffer */
#define GET_BYTE(i) (*i++)
/*! \brief Get long (4-byte value) from input buffer */
/* #define GET_LONG(i) (GET_BYTE(i) | (GET_BYTE(i) << 8) | (GET_BYTE(i) << 16) | (GET_BYTE(i) << 24))    - Doesn't work correctly on Linux */
#define GET_LONG(i) ((*i) | (*(i+1) << 8) | (*(i+2) << 16) | (*(i+3) << 24))
/*! \brief Check IPP DC status. If not ippStsNoErr - complain */
#define CHECK_STATUS(status) if (status != ippStsNoErr) print_ipp_status(status);

/* MS-specific definitions */
#if defined (_WIN32) && _MSC_VER >= 1400 && !defined(_AMD64_)
    #define UNLINK(file) _unlink(file)
    #define FILENO(file) _fileno(file)
    #define ISATTY(file) _isatty(file)
    #define SETMODE(file, mode) _setmode(file, mode)
    #define SYSOPEN _sopen_s
    #define SYSCLOSE _close
    #define SYSREAD _read
    #define SYSWRITE _write
    #define SYSLSEEK _lseek
    #define SYSFSTAT _fstat64i32
    #define SYSTELL _tell
#else
    #define UNLINK(file) unlink(file)
    #define FILENO(file) fileno(file)
    #define ISATTY(file) isatty(file)
    #define SETMODE(file, mode) setmode(file, mode)
    #define SYSOPEN open
    #define SYSCLOSE close
    #define SYSREAD read
    #define SYSWRITE write
    #define SYSLSEEK lseek
    #define SYSFSTAT fstat
#ifdef _WIN32
    #define SYSTELL tell
#else
    #define SYSTELL(f) lseek(f, 0, SEEK_CUR)
#endif
#endif
#ifdef _WIN32
#if _MSC_VER >= 1400  && !defined(_AMD64_)
    #define READ_MODE_BITS  (_O_RDONLY | _O_SEQUENTIAL | _O_BINARY)
    #define WRITE_MODE_BITS (_O_CREAT | _O_BINARY | _O_RDWR | _O_SEQUENTIAL | _O_TRUNC)
#else
    #define READ_MODE_BITS  (O_RDONLY | O_SEQUENTIAL | O_BINARY)
    #define WRITE_MODE_BITS (O_CREAT | O_BINARY | O_RDWR | O_SEQUENTIAL | O_TRUNC)
#endif
    #define SYSUNMAP(p,l) UnmapViewOfFile(p)
    #define SYSCLOSE_FD_OR_HANDLE(fd, h) CloseHandle(h)
#else
    #define READ_MODE_BITS  (O_RDONLY)
    #define WRITE_MODE_BITS (O_CREAT | O_RDWR | O_TRUNC)
    #define SYSUNMAP(p,l) munmap(p, l)
    #define SYSCLOSE_FD_OR_HANDLE(fd, h) SYSCLOSE(fd)
#endif

/*! \brief Choose minimal of two values */
#define MINSIZE(a, b) ((a<b)? a: b)

#endif
