/*
//
//                  INTEL CORPORATION PROPRIETARY INFORMATION
//     This software is supplied under the terms of a license agreement or
//     nondisclosure agreement with Intel Corporation and may not be copied
//     or disclosed except in accordance with the terms of that agreement.
//          Copyright(c) 2007-2008 Intel Corporation. All Rights Reserved.
//
//     Intel(R) Integrated Performance Primitives Data Compression Domain-based GZIP
//
//  By downloading and installing this sample, you hereby agree that the
//  accompanying Materials are being provided to you under the terms and
//  conditions of the End User License Agreement for the Intel(R) Integrated
//  Performance Primitives product previously accepted by you. Please refer
//  to the file ippEULA.rtf located in the root directory of your Intel(R) IPP
//  product installation for more information.
//
*/
#include "ipp_gzip.h"

    #include <fcntl.h>
    #ifdef _WIN32
        #include <io.h>
    #include <sys\utime.h>
#if _MSC_VER >= 1400 && !defined(_AMD64_)
#include <share.h>
#endif
extern SYSTEM_INFO sysInfo;
#else
#include <unistd.h>
    #include <sys/time.h>
#include <string.h>
#include <stdlib.h>
#endif
/*
 map_view_file - map portion of file to virtual memory of our task
*/
static int map_view_file(GZIP_FILE *file, GZIP_FILE_MODE mode, size_t initial_size)
{
#ifndef _WIN32
size_t sys_page_size;
#endif

    file->mode = mode;
#ifdef _WIN32
    file->allocation_granularity = sysInfo.dwAllocationGranularity;     /* It must be 64K but anyway...*/
    if (mode == GZIP_FILE_READ)
        file->hMap = CreateFileMapping(file->hFile, NULL, PAGE_READONLY, 0, 0, NULL);
    else
        file->hMap = CreateFileMapping(file->hFile, NULL, PAGE_READWRITE, 0, (DWORD)(initial_size > OUTBUF_SIZE?initial_size : OUTBUF_SIZE), NULL);
    if (file->hMap == NULL) {
        error_system("Cannot create map object");
        return 0;
    }
    if (mode == GZIP_FILE_READ) {
        file->valid_bytes = MINSIZE(file->size,INBUF_SIZE);
        file->pview_pos = file->pmap = MapViewOfFile(file->hMap, FILE_MAP_READ, 0, 0, file->valid_bytes);
    } else {
        file->valid_bytes = MINSIZE(initial_size, OUTBUF_SIZE);
        file->pview_pos = file->pmap = MapViewOfFile(file->hMap, FILE_MAP_ALL_ACCESS, 0, 0, file->valid_bytes);
    }
    if (file->pmap == NULL) {
        error_system("Cannot map view of file");
        CloseHandle(file->hMap);
        return 0;
    }
#else
    ftruncate(file->fd, initial_size);
    initial_size = initial_size;
    sys_page_size = (size_t)sysconf(_SC_PAGESIZE);
    if (mode == GZIP_FILE_READ) {
        file->maplen = file->allocation_granularity = 
            (INBUF_SIZE / sys_page_size) * sys_page_size;   /* Round-up to SC_PAGESIZE boundary */
        file->valid_bytes = MINSIZE(file->maplen, file->size);
        file->pview_pos = file->pmap = mmap(NULL, file->valid_bytes, PROT_READ, MAP_SHARED, file->fd, file->fileOffset);
    } else {
        file->maplen = file->allocation_granularity = 
            (OUTBUF_SIZE / sys_page_size) * sys_page_size;  /* Round-up to SC_PAGESIZE boundary */
        file->valid_bytes = MINSIZE(initial_size, file->maplen);
        file->maplen = file->valid_bytes;
        file->pview_pos = file->pmap = mmap(NULL, file->valid_bytes, PROT_READ | PROT_WRITE, MAP_SHARED, file->fd, file->fileOffset);
    }
    if (file->pmap == (void*)-1) {
        error_system("Cannot map file");
        return 0;
    }
#endif
    return 1;
}
GZIP_FILE* ZOPEN(char *filename, GZIP_FILE_MODE mode, size_t initial_size)
{
    GZIP_FILE *new_ptr;
#if _MSC_VER >= 1400  && !defined(_AMD64_)
    struct _stat istat;
#else
    struct stat istat;
#endif

    new_ptr = (GZIP_FILE*)zalloc(sizeof(GZIP_FILE));
    new_ptr->bytes_out = 0;
    new_ptr->bytes_in = 0;
    new_ptr->eof_condition = 0;
    new_ptr->mode = mode;
    new_ptr->uncompressed_size = 0;
    new_ptr->pmap = NULL;
    if (mode == GZIP_FILE_READ) {
#if _MSC_VER >= 1400  && !defined(_AMD64_)
        _stat(filename, &istat);
#else
        stat(filename, &istat);
#endif
        new_ptr->size = istat.st_size;
        if (opt_io_read_method > METHOD_LOW) {       /* =1 means we want to use mmap for reading */
            #ifdef _WIN32           /* Specific for Windows */
                new_ptr->hFile = CreateFile(filename, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL | FILE_FLAG_SEQUENTIAL_SCAN, NULL);
                if (new_ptr->hFile == INVALID_HANDLE_VALUE) {
                    error_file("Cannot create file");
                    zfree(new_ptr);
                    return NULL;
                }
                new_ptr->fileOffset = 0;                /* Map view of file */
                new_ptr->curFilePosition = 0;
                if (map_view_file(new_ptr, mode, initial_size) == 0) {
                    CloseHandle(new_ptr->hFile);
                    zfree(new_ptr);
                    return NULL;
                }
            #else                   /* Specific for Linux */
                if ((new_ptr->fd = SYSOPEN(filename, READ_MODE_BITS, S_IWRITE | S_IREAD)) < 0) {
                    zfree(new_ptr);
                    return NULL;
                }
                new_ptr->fileOffset = 0;
                new_ptr->curFilePosition = 0;
                if (map_view_file(new_ptr, mode, initial_size) == 0) {
                    close(new_ptr->fd);
                    zfree(new_ptr);
                    return NULL;
                }
#endif
            new_ptr->uncompressed_size = 0;
        } else {                            /* opt_io_read_method==0 means we want to use low-level */
#if _MSC_VER >= 1400 && !defined(_AMD64_)
            if (SYSOPEN(&new_ptr->fd, filename, READ_MODE_BITS, _SH_DENYNO, _S_IWRITE | _S_IREAD)) {
#else
            if ((new_ptr->fd = SYSOPEN(filename, READ_MODE_BITS, S_IWRITE | S_IREAD)) == -1) {
#endif
        zfree(new_ptr);
        return NULL;
    }
#if _MSC_VER >= 1400  && !defined(_AMD64_)
            _stat(filename, &istat);
#else
            stat(filename, &istat);
#endif
            new_ptr->size = istat.st_size;
        }
        initial_size = initial_size;    /* To get rid of "not used" compiler message */
    return new_ptr;
    } else {        /* GZIP_FILE_WRITE */
        new_ptr->size = 0;
        if (opt_io_write_method > METHOD_LOW && !opt_to_stdout) {      /* mmap */
#ifdef _WIN32           /* Specific for Windows */
                new_ptr->hFile = CreateFile(filename, FILE_ALL_ACCESS, FILE_SHARE_WRITE, NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL | FILE_FLAG_SEQUENTIAL_SCAN, NULL);
                if (new_ptr->hFile == INVALID_HANDLE_VALUE) {
                    error_file("Cannot create file");
                    zfree(new_ptr);
                    return NULL;
                }
#else                   /* Specific for Linux */
                if ((new_ptr->fd = SYSOPEN(filename, WRITE_MODE_BITS, S_IWRITE | S_IREAD)) < 0) {
                    zfree(new_ptr);
                    return NULL;
                }
#endif
            new_ptr->fileOffset = 0;                /* Map view of file */
            new_ptr->curFilePosition = 0;
            if (map_view_file(new_ptr, mode, initial_size) == 0) {
                SYSCLOSE_FD_OR_HANDLE(new_ptr->fd, new_ptr->hFile);
                zfree(new_ptr);
                return NULL;
            }
            new_ptr->uncompressed_size = 0;
        } else {                            /* low-level writes */
#if _MSC_VER >= 1400 && !defined(_AMD64_)
            if (SYSOPEN(&new_ptr->fd, filename, WRITE_MODE_BITS, _SH_DENYNO, _S_IWRITE | _S_IREAD)) {
#else
            if ((new_ptr->fd = SYSOPEN(filename, WRITE_MODE_BITS, S_IWRITE | S_IREAD)) == -1) {
#endif
                zfree(new_ptr);
                return NULL;
            }
            #ifdef _WIN32
                new_ptr->hFile = (HANDLE)_get_osfhandle(new_ptr->fd);   /* Restore file handle for possible mmap operations */
            #endif
        }   /* if (write_method)*/
        return new_ptr;
    }   /* if (read or write)*/
}
/**
 * return GZIP_FILE structure for terminal output device
 * 
 * @return GZIP_FILE*
 */
GZIP_FILE* zopen_stdout()
{
    GZIP_FILE *new_ptr;
    new_ptr = (GZIP_FILE*)zalloc(sizeof(GZIP_FILE));
    new_ptr->bytes_in = 0;
    new_ptr->bytes_out = 0;
    new_ptr->mode = GZIP_FILE_WRITE;
    new_ptr->fd = FILENO(stdout);
#ifdef _WIN32
#if _MSC_VER >= 1400  && !defined(_AMD64_)
    SETMODE(new_ptr->fd, _O_BINARY);
#else
    SETMODE(new_ptr->fd, O_BINARY);
#endif
#endif
    return new_ptr;
}
/**
 * returns GZIP_FILE pointer to stdin stream
 * 
 * @return GZIP_FILE*
 */
GZIP_FILE* zopen_stdin()
{
    GZIP_FILE *new_ptr;
    new_ptr = (GZIP_FILE*)zalloc(sizeof(GZIP_FILE));
    new_ptr->bytes_in = 0;
    new_ptr->bytes_out = 0;
    new_ptr->eof_condition = 0;
    new_ptr->fd = FILENO(stdin);
#ifdef _WIN32
#if _MSC_VER >= 1400  && !defined(_AMD64_)
    SETMODE(new_ptr->fd, _O_BINARY);
#else
    SETMODE(new_ptr->fd, O_BINARY);
#endif
#endif
    return new_ptr;
}

int zclose(GZIP_FILE *file)
{
    int close_status;

    if ((file->mode == GZIP_FILE_READ && (opt_io_read_method == METHOD_LOW || from_stdin)) ||
        (file->mode == GZIP_FILE_WRITE && (opt_io_write_method == METHOD_LOW || opt_to_stdout))) {
        close_status = SYSCLOSE(file->fd);
    zfree(file);
    return close_status;
    }
    /* Otherwise file i/o method was METHOD_MAP */
    if (file->pmap != NULL) {
        SYSUNMAP(file->pmap, file->maplen);
        file->pmap = NULL;
    }
    #ifdef _WIN32
        CloseHandle(file->hMap);
#endif
    if (file->mode == GZIP_FILE_WRITE)
        ztruncate(file);
    SYSCLOSE_FD_OR_HANDLE(file->fd, file->hFile);
    zfree(file);
    return 1;
}
/*--------------------------------------------------------------------------------*/
/**
 * 
 * zread - read specified number of bytes into specified buffer.
 * Return number of bytes read.
 * 
 */
size_t zread(GZIP_FILE *file, UBYTE *buffer, size_t size)
{
    size_t to_read;

    to_read = size;
    if (opt_io_read_method == METHOD_LOW || from_stdin) {
        int incnt;
        while ((incnt=SYSREAD(file->fd, buffer, (unsigned int)to_read)) > 0) {
        to_read -= incnt;
        if (to_read == 0)
            break;
            buffer += incnt;
        }
        if (incnt == -1) {
            perror("Cannot read from file");
            return 0;
    }
    file->eof_condition = (incnt == 0)? 1 : 0;
    file->bytes_in += (size-to_read);
        return size-to_read;
    } else {    /* METHOD_MAP */
        size_t byte_count = 0;
        if (file->mode == GZIP_FILE_WRITE) {    /* This is for re-reading of tmp files, written by low/level, using mmap */
            if (file->pmap == NULL) {           /* No mapping was done yet */
                file->fileOffset = 0;                /* Map view of file */
                if (map_view_file(file, GZIP_FILE_READ, file->size) == 0) {
                    fprintf(stderr, "Re-mapping of low/level file failed\n");
                    exit(1);                    /* Nothing to do here */
                }
                file->curFilePosition = 0;
                file->uncompressed_size = 0;
            }
        }
        while (to_read > 0) {
            if (file->valid_bytes > 0) {
                if (to_read <= file->valid_bytes) {
                    memcpy(buffer, file->pview_pos, to_read);
                    file->pview_pos += to_read;
                    file->valid_bytes -= to_read;
                    byte_count = byte_count + to_read;
                    file->bytes_in += byte_count;
                    file->curFilePosition = file->fileOffset + (file->pview_pos - file->pmap);
                    return byte_count;
                }
                memcpy(buffer, file->pview_pos, file->valid_bytes);
                file->pview_pos += file->valid_bytes;
                buffer += file->valid_bytes;
                byte_count = byte_count + file->valid_bytes;
                to_read = to_read - file->valid_bytes;
                if (file->eof_condition) {
                    file->bytes_in += byte_count;
                    file->curFilePosition = file->fileOffset + (file->pview_pos - file->pmap);
                    return byte_count;
                }
            }
            /* Re-map file */
            file->fileOffset += (DWORD)INBUF_SIZE;
            if (file->fileOffset >= file->size) {           /* File size is multiple of AllocationGranularity */
                file->eof_condition = 1;
                file->bytes_in += byte_count;
                file->curFilePosition = file->fileOffset + ((UBYTE*)file->pview_pos - (UBYTE*)file->pmap);
                return byte_count;
            }
            SYSUNMAP(file->pmap, file->maplen);
    #ifdef _WIN32
            file->pview_pos = file->pmap = MapViewOfFile(file->hMap, FILE_MAP_READ, 0, (DWORD)file->fileOffset, 
                (file->fileOffset + INBUF_SIZE) > file->size? file->size - file->fileOffset : INBUF_SIZE);
            if (file->pmap == NULL) {
                error_system("Cannot remap file");
                exit(1);
            }
    #else
            file->maplen = ((file->fileOffset + file->allocation_granularity) > file->size)? 
                                file->size % file->allocation_granularity : file->allocation_granularity;
            file->pview_pos = file->pmap = mmap(NULL, file->maplen, PROT_READ, MAP_SHARED, file->fd, file->fileOffset);
            if (file->pmap == (void*)-1) {
                error_system("Cannot remap file");
                exit(1);
            }
#endif
            if (file->fileOffset + INBUF_SIZE >= file->size)
                file->valid_bytes = file->size % INBUF_SIZE;
            else
                file->valid_bytes = INBUF_SIZE;
        } /* while (copy_count > 0) */
        file->bytes_in += byte_count;
        file->curFilePosition = file->fileOffset + (file->pview_pos - file->pmap);
        return byte_count;
}
}
/*--------------------------------------------------------------------------------*/
/**
 * zwrite - write specified number of bytes from specified
 * buffer address
 * 
 * @param file
 * @param buffer
 * @param size
 * 
 * @return size_t
 */
size_t zwrite(GZIP_FILE *file, UBYTE *buffer, size_t size)
{
    if (opt_io_write_method == METHOD_LOW || opt_to_stdout) {
    size_t outcnt;
#if (defined _WIN32) && (_MSC_VER >= 1400)  && !defined(_AMD64_)
        struct _stat64i32 ostat;
#else
        struct stat ostat;
#endif
        outcnt = SYSWRITE(file->fd, buffer, (unsigned int)size);
    file->bytes_out += outcnt;
        SYSFSTAT(file->fd, &ostat);
        file->size = ostat.st_size;
        return outcnt;
    } else {    /* MODE_MMAP */
        size_t copy_count = size;
        size_t byte_count = 0;

        while (copy_count > 0) {
            if (copy_count <= file->valid_bytes) {       /* We have enough space in view window */
                memcpy(file->pview_pos, buffer, copy_count);
                file->pview_pos += copy_count;
                file->valid_bytes -= copy_count;
                byte_count = byte_count + copy_count;
                file->bytes_out += byte_count;
                file->curFilePosition = file->fileOffset + (file->pview_pos - file->pmap);
                if (file->size < file->curFilePosition)
                    file->size = file->curFilePosition;
                return byte_count;
            } else {                                    /* Not enough space in view window */
                memcpy(file->pview_pos, buffer, file->valid_bytes);
                buffer += file->valid_bytes;
                copy_count = copy_count - file->valid_bytes;
                byte_count = byte_count + file->valid_bytes;
                /* Remap view window */
                file->fileOffset += (DWORD)OUTBUF_SIZE;
                SYSUNMAP(file->pmap, file->maplen);
#ifdef _WIN32
                file->valid_bytes = MINSIZE(file->uncompressed_size-file->fileOffset, OUTBUF_SIZE);
                file->pmap = file->pview_pos = (UBYTE*)MapViewOfFile(file->hMap, FILE_MAP_ALL_ACCESS, 0, (DWORD)file->fileOffset, file->valid_bytes);
                if (file->pmap == NULL) {
                    error_system("Cannot remap file");
                    return 0;
                }
#else
                file->maplen = file->valid_bytes = MINSIZE(file->uncompressed_size-file->fileOffset, OUTBUF_SIZE);
                file->pmap = file->pview_pos = (UBYTE*)mmap(NULL, file->maplen, PROT_WRITE, MAP_SHARED, file->fd, file->fileOffset);
                if (file->pmap == (void*)-1) {
                    error_system("Cannot remap file");
                    return 0;
                }
#endif
            } /* if (copy <= valid) */
        } /* while(copy_count>0)*/
        file->bytes_out += byte_count;
        file->curFilePosition = file->fileOffset + ((UBYTE*)file->pview_pos - (UBYTE*)file->pmap);
        if (file->size < file->curFilePosition)
            file->size = file->curFilePosition;
        return byte_count;
}
}
/*--------------------------------------------------------------------------------*/
int zseek(GZIP_FILE *file, long off, int origin)
{
    if (opt_to_stdout || (file->mode == GZIP_FILE_READ && opt_io_read_method == METHOD_LOW) ||
        (file->mode == GZIP_FILE_WRITE && opt_io_write_method == METHOD_LOW)) {
        return (SYSLSEEK(file->fd, off, origin) == off)? 0 : 1;         /* fseek compatible return */
}
    else {
        long offset_from_beg;

        if (origin == SEEK_END)
            offset_from_beg = (long)(file->size - abs(off));
        else
            offset_from_beg = off;
        file->fileOffset = (DWORD)(offset_from_beg - offset_from_beg % ((file->mode==GZIP_FILE_READ)? INBUF_SIZE : OUTBUF_SIZE));
#ifdef _WIN32
        if (file->hMap == INVALID_HANDLE_VALUE) {       /* Means, that we closed the view. Re-create view */
            if (map_view_file(file, GZIP_FILE_READ, file->size) == 0) {
                fprintf(stderr, "Re-mapping of low/level file failed\n");
                exit(1);                    /* Nothing to do here */
            }
        } else {
            size_t buf_size;
            if (file->pmap != NULL) {
                UnmapViewOfFile(file->pmap);
                file->pmap = NULL;
            }
            if (file->mode == GZIP_FILE_READ) {
                buf_size = MINSIZE(file->size - file->fileOffset, INBUF_SIZE);
                file->pmap = MapViewOfFile(file->hMap, FILE_MAP_READ, 0, (DWORD)file->fileOffset, buf_size);
            } else {
                buf_size = MINSIZE(file->uncompressed_size, OUTBUF_SIZE);
                file->pmap = MapViewOfFile(file->hMap, FILE_MAP_ALL_ACCESS, 0, (DWORD)file->fileOffset, buf_size);
            }
            if (file->pmap == NULL) {
                error_system("Cannot remap file");
                return 1;
            }
            file->pview_pos = file->pmap + offset_from_beg % buf_size;
            if (file->size < (offset_from_beg + buf_size))
                file->valid_bytes = file->size - offset_from_beg;
            else
                file->valid_bytes = buf_size - (offset_from_beg % buf_size);
        }
#else
        munmap(file->pmap, file->maplen);
        file->maplen = file->allocation_granularity;
        file->pmap = file->pview_pos = (UBYTE*)mmap(NULL, file->maplen, (file->mode==GZIP_FILE_READ)? PROT_READ : PROT_READ | PROT_WRITE, 
                                                    MAP_SHARED, file->fd, file->fileOffset);
        if (file->pmap == (void*)-1) {
            error_system("Cannot remap file");
            return 0;
        }
        file->pview_pos = file->pmap + offset_from_beg % file->allocation_granularity;
        if (file->size < (offset_from_beg + file->allocation_granularity))
            file->valid_bytes = file->size - offset_from_beg;
        else
            file->valid_bytes = file->allocation_granularity - (offset_from_beg % file->allocation_granularity);
#endif
        return 0;
}
}
/*--------------------------------------------------------------------------------*/
/**
 * zeof - return 1 if EOF, 0 otherwise
 * @param file
 * 
 * @return int
 */
int zeof(GZIP_FILE *file)
{
    return file->eof_condition;
}
/*--------------------------------------------------------------------------------*/
long ztell(GZIP_FILE *file)
{
    if (opt_to_stdout || (file->mode == GZIP_FILE_READ && opt_io_read_method == METHOD_LOW) ||
        (file->mode == GZIP_FILE_WRITE && opt_io_write_method == METHOD_LOW)) {
        return SYSTELL(file->fd);
}
    else
        return (long)file->curFilePosition;
}
/*--------------------------------------------------------------------------------*/
void zcopytime(char *infilename, char *outfilename, long timestamp)
{
#if _MSC_VER >= 1400  && !defined(_AMD64_)
    struct _stat ifstat;
#else
    struct stat ifstat;
#endif
#ifdef _WIN32
    struct _utimbuf utim;
#else
    struct timeval file_times[2];
#endif
    if (infilename == NULL) {       /* Inflating from stdin */
#if _MSC_VER >= 1400  && !defined(_AMD64_)
        if (_fstat(FILENO(stdin), &ifstat) != 0) {
#else
        if (fstat(FILENO(stdin), &ifstat) != 0) {
#endif
            error_file("stdin");
            return;
        }
    } else {
#if _MSC_VER >= 1400  && !defined(_AMD64_)
        if (_stat(infilename, &ifstat)) {
#else
        if (stat(infilename, &ifstat)) {
#endif
            error_file(infilename);
            return;             /* Couldn't get input file information */
        }
    }
    if (!opt_notime && timestamp)
        ifstat.st_mtime = timestamp;
#ifdef _WIN32
    utim.actime = ifstat.st_atime;
    utim.modtime = ifstat.st_mtime;
    _utime(outfilename, &utim);
#else
    file_times[0].tv_sec = ifstat.st_atime;
    file_times[1].tv_sec = ifstat.st_mtime;
    utimes(outfilename, file_times);
#endif
}
/*--------------------------------------------------------------------------------*/
/**
 * Set EOF at current file position.
 * @param file
 */
void ztruncate(GZIP_FILE *file)
{
    if (opt_io_write_method == METHOD_LOW)
        return;     /* We don't need this on low/level i/o */
    if (file->pmap != NULL) {
        SYSUNMAP(file->pmap, file->maplen);
#ifdef _WIN32
        CloseHandle(file->hMap);            /* We don't need view window anymore */
        file->hMap = INVALID_HANDLE_VALUE;
#endif
        file->pmap = NULL;
    }
#ifdef _WIN32
    SetFilePointer(file->hFile, (LONG)file->size, 0, FILE_BEGIN);
    SetEndOfFile(file->hFile);
#else
    ftruncate(file->fd, file->size);
#endif
}

/*--------------------------------------------------------------------------------*/
/**
 * Returns uncompressed file size. This is for .gz files, where
 * uncompressed files size is written at the end of file
 * @param file
 * 
 * @return DWORD
 */
DWORD zsizeuncompressed(GZIP_FILE *file)
{
    int file_tail[2];
    long cur_position;

    if (file->uncompressed_size != 0) {
        return (DWORD)file->uncompressed_size;
    }
    /* Save current file position */
    cur_position = ztell(file);
    /* Read tail 8 bytes of file */
    zseek(file, -8, SEEK_END);
    zread(file, (UBYTE*)file_tail, 8);
    file->uncompressed_size = file_tail[1];
    /* Restore current position */
    zseek(file, cur_position, SEEK_SET);
    return file_tail[1];
}

/*--------------------------------------------------------------------------------*/
/**
 * zextend - extend output file size
 * 
 */
BOOL zextend(GZIP_FILE *file, DWORD target_size)
{
#ifdef _WIN32
    BOOL retVal;
    if (SetFilePointer(file->hFile, target_size, 0, FILE_END) != target_size) {
        return FALSE;
    }
    retVal = SetEndOfFile(file->hFile);
    SetFilePointer(file->hFile, 0, 0, FILE_BEGIN);
    return retVal;
#else
    return (BOOL)(ftruncate(file->fd, target_size)==0);
#endif
}
/*--------------------------------------------------------------------------------*/
size_t zgetsize(GZIP_FILE *file)
{
    return file->size;
}
/*--------------------------------------------------------------------------------*/
#ifdef GZIP_PTHREADS
static pthread_mutex_t cur_position_mutex = PTHREAD_MUTEX_INITIALIZER;
#endif
GZIP_VIEW* zopenview(GZIP_FILE* file, DWORD offset)
{
    GZIP_VIEW *new_view;

    new_view = zalloc(sizeof(GZIP_VIEW));
    new_view->size = 0;
    new_view->fileOffset = offset - (offset % OUTBUF_SIZE);
    new_view->file_ptr = file;
#ifdef _WIN32
    new_view->pmap = (UBYTE*)MapViewOfFile(file->hMap, FILE_MAP_ALL_ACCESS, 0, (DWORD)new_view->fileOffset, 
        MINSIZE(file->uncompressed_size-new_view->fileOffset, OUTBUF_SIZE));
    if (new_view->pmap == NULL) {
        error_system("Cannot create view of file");
        zfree(new_view);
        return NULL;
    }
#else
    new_view->maplen = new_view->file_ptr->allocation_granularity;
    new_view->pmap = new_view->pview_pos = mmap(NULL, new_view->maplen, PROT_WRITE, MAP_SHARED, 
                                                    new_view->file_ptr->fd, new_view->fileOffset);
    if (new_view->pmap == (void*)-1) {
        error_system("Cannot map file");
        zfree(new_view);
        return 0;
    }
#endif
    new_view->bytes_out = 0;
    new_view->pview_pos = new_view->pmap + (offset % OUTBUF_SIZE);
    new_view->valid_bytes = MINSIZE(file->uncompressed_size-new_view->fileOffset, OUTBUF_SIZE) - (offset % OUTBUF_SIZE);
#ifdef GZIP_PTHREADS
    /* For PTHREAD version we need mutex to update output file size from within different threads */
    pthread_mutex_init(&cur_position_mutex, NULL);
#endif
    return new_view;
}
/*--------------------------------------------------------------------------------*/
size_t zwriteview(GZIP_VIEW *view, UBYTE *buffer, size_t size)
{
    size_t copy_count = size;
    size_t byte_count = 0;

    while (copy_count > 0) {
        if (copy_count <= view->valid_bytes) {       /* We have enough space in view window */
            memcpy(view->pview_pos, buffer, copy_count);
            view->pview_pos += copy_count;
            view->valid_bytes -= copy_count;
            byte_count = byte_count + copy_count;
            view->bytes_out += byte_count;
            view->curFilePosition = view->fileOffset + (view->pview_pos - view->pmap);
            if (view->size < view->curFilePosition)
                view->size = view->curFilePosition;
            return byte_count;
        } else {                                    /* Not enough space in view window */
            memcpy(view->pview_pos, buffer, view->valid_bytes);
            buffer += view->valid_bytes;
            copy_count = copy_count - view->valid_bytes;
            byte_count = byte_count + view->valid_bytes;
            /* Remap view window */
            view->fileOffset += OUTBUF_SIZE;
#ifdef _WIN32
            UnmapViewOfFile(view->pmap);
            view->valid_bytes = MINSIZE(view->file_ptr->uncompressed_size - view->fileOffset, OUTBUF_SIZE);
            view->pmap = view->pview_pos = (UBYTE*)MapViewOfFile(view->file_ptr->hMap, FILE_MAP_ALL_ACCESS, 0, (DWORD)view->fileOffset, 
                view->valid_bytes);
            if (view->pmap == NULL) {
                error_system("Cannot remap file");
                return 0;
            }
#else
            munmap(view->pmap, view->maplen);
            view->maplen = view->file_ptr->allocation_granularity;
            view->pmap = view->pview_pos = mmap(NULL, view->maplen, PROT_WRITE, MAP_SHARED, view->file_ptr->fd, view->fileOffset);
            if (view->pmap == (void*)-1) {
                error_system("Cannot remap file");
                return 0;
            }
            view->valid_bytes = MINSIZE(view->file_ptr->uncompressed_size - view->fileOffset, OUTBUF_SIZE);
#endif
        } /* if (copy <= valid) */
    } /* while(copy_count>0)*/
    view->bytes_out += byte_count;
    view->curFilePosition = view->fileOffset + (view->pview_pos - view->pmap);
    if (view->size < view->curFilePosition)
        view->size = view->curFilePosition;
    return byte_count;
}
/*--------------------------------------------------------------------------------*/
void zcloseview(GZIP_VIEW *view)
{
    if (view->pmap != NULL) {
        #ifdef _WIN32
            UnmapViewOfFile((LPCVOID)view->pmap);
            view->pmap = NULL;
        #else
        #endif
    }
#ifdef GZIP_PTHREADS
    pthread_mutex_lock(&cur_position_mutex);            /* Initialize critical section */
    if (view->file_ptr->size < view->curFilePosition) {
        view->file_ptr->size = view->curFilePosition;
        view->file_ptr->bytes_out = view->curFilePosition;
    }
    pthread_mutex_unlock(&cur_position_mutex);
#else
#if defined(_OPENMP) && defined(GZIP_PARALLEL)
#pragma omp critical
#endif
    {
        if (view->file_ptr->size < view->curFilePosition) {
            view->file_ptr->size = view->curFilePosition;
            view->file_ptr->bytes_out = view->curFilePosition;
        }
    }
#endif
    zfree(view);
}

